import os
from warnings import warn
import sys
from pmc_automation_tools.driver.common import (
    PlexDriver,
    PlexElement,
    VISIBLE,
    INVISIBLE,
    CLICKABLE,
    EXISTS,
    SIGNON_URL_PARTS
    )
from selenium.common.exceptions import (
    TimeoutException,
    StaleElementReferenceException,
    NoSuchElementException
    )
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import WebDriverWait, Select
from selenium.webdriver.support import expected_conditions as EC
from pmc_automation_tools.common.exceptions import (
    UpdateError,
    NoRecordError,
    LoginError,
    PlexAutomateError
)
import time
from tkinter import filedialog
from tkinter import messagebox
import json
import csv
from pmc_automation_tools.common.utils import (
    get_case_insensitive_key_value
)

PLEX_GEARS_SELECTOR = (By.ID, '__WAITMESSAGE_CONTAINER')
PCN_SQL = '''Please create the pcn.json file by running the following SQL report in Plex and save it as a csv file.

SELECT
 P.Plexus_Customer_No
, P.Plexus_Customer_Name
FROM Plexus_Control_v_Customer_Group_Member P

Press OK to select the csv file.'''
class ClassicDriver(PlexDriver):
    def __init__(self, *args, **kwargs):
        super().__init__(environment='classic', *args, **kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.pcn_file_path = kwargs.get('pcn_file_path', os.path.join('resources', 'pcn.json'))
        self._pcn_file_check()

    def _pcn_file_check(self):
        while not os.path.exists(self.pcn_file_path):
            confirm = messagebox.askokcancel(title='Classic PCN reference file is missing',
                                            message=PCN_SQL)
            if not confirm:
                messagebox.showinfo(title='User Selected Cancel',
                                    message='The program will now close.')
                sys.exit("Process quit by user")
            self.file_path = filedialog.askopenfilename()
            if self.file_path:
                self._csv_to_json(self.file_path)
        self.pcn_dict = {}
        with open(self.pcn_file_path, 'r', encoding='utf-8') as pcn_config:
            self.pcn_dict = json.load(pcn_config)

    def _csv_to_json(self, csv_file):
        """Function to take a csv file from Plex and create a
        
        PCN JSON file that can be used to log into specific PCNs
            if the user has multiple PCN access.
        
        This should only be called on initialization if
            the pcn.json file does not exist yet.
        
        Only required for classic logins.

        Args:
            csv_file (str): path to the csv file generated from the SQL development environment.
        """
        _pcn_dict = {}
        with open(csv_file, 'r', encoding='utf-8-sig') as c:
            r = csv.DictReader(c)
            for row in r:
                if not row:
                    continue
                _pcn_dict[get_case_insensitive_key_value(row, 'plexus_customer_no')] = get_case_insensitive_key_value(row, 'plexus_customer_name')
        if not os.path.exists('resources'):
            os.mkdir('resources')
        with open(os.path.join('resources', 'pcn.json'), 'w+', encoding='utf-8') as j:
            j.write(json.dumps(_pcn_dict, indent=4, ensure_ascii=False))

    def wait_for_element(self, selector, driver=None, timeout=15, type=VISIBLE, ignore_exception=False):
        return super().wait_for_element(selector, driver=driver, timeout=timeout, type=type, ignore_exception=ignore_exception, element_class=ClassicPlexElement)

    def wait_for_gears(self, loading_timeout=10):
        super().wait_for_gears(PLEX_GEARS_SELECTOR, loading_timeout)

    def click_button(self, button_text:str, driver:'ClassicPlexElement'=None):
        """Click on a button.

        Classic buttons have two different types of buttons.
            - buttons using a/span tags
            - buttons made of ul/li tags

        Args:
            button_text (str): Text matching the button to click
            driver (ClassicPlexElement, optional): The WebDriver to use as a root for searching for the button. Defaults to None.
        """
        driver = driver or self.driver
        # Elements with <a><span> button structure
        a_buttons = driver.find_elements(By.CLASS_NAME, 'Button')
        # Elements with <ul><li> button structure
        ul_buttons = driver.find_elements(By.CLASS_NAME, 'button')
        buttons = a_buttons + ul_buttons
        for b in buttons:
            if b.get_property('textContent') == button_text:
                self.debug_logger.debug(f'Button found with matching text: {button_text}')
                b.click()
                break


    def login(self, username, password, company_code, pcn, test_db=True, headless=False):
        """Log in to Plex

        Args:
            username (str): Plex username
            password (str): Plex password
            company_code (str): Plex company code (No longer relevant)
            pcn (str): PCN number
            test_db (bool, optional): Determines if logging in to the test database. Defaults to True.
            headless (bool, optional): Run in headless mode. Defaults to False.

        Returns:
            tuple: WebDriver and combine url for navigating to other screens.
        """
        self._set_login_vars()
        super().login(username, password, company_code, pcn, test_db, headless)
        self._classic_popup_handle()
        self._login_validate()
        self.pcn_switch(self.pcn)
        self.token_get()
        self.first_login = False
        return (self.driver, self.url_comb)
    

    def _set_login_vars(self):
        self.plex_main = 'plexonline.com'
        self.plex_prod = 'www.'
        self.plex_test = 'test.'
        self.sso = '/signon'
        super()._set_login_vars()

    def _classic_popup_handle(self):
        """Deal with the browser popup that occurs after logging in to Plex classic.

        Raises:
            LoginError: If there is any credential issues, the window will not appear.
        """
        main_window_handle = self.driver.current_window_handle
        signin_window_handle = None
        timeout_signin = 30
        timeout_start = time.time()
        while time.time() < timeout_start + timeout_signin:
            self.debug_logger.debug(f'Searching for classic signin window')
            window_handles = self.driver.window_handles
            if len(window_handles) > 1:
                for handle in window_handles:
                    if handle != main_window_handle:
                        self.debug_logger.debug(f'Found classic login window')
                        signin_window_handle = handle
                        break
                if signin_window_handle:
                    break
                time.sleep(1)
        if not signin_window_handle:
            raise LoginError('Failed to find Plex signon window. Please validate login credentials and try again.')
        self.driver.switch_to.window(signin_window_handle)


    def _login_validate(self):
        url = self.driver.current_url
        if not any(url_part in url.upper() for url_part in SIGNON_URL_PARTS):
            raise LoginError('Login page not detected. Please validate login credentials and try again.')


    def _pcn_switch(self, pcn):
        _pcn_name = self.pcn_dict.get(pcn, None)
        if not _pcn_name:
            raise LoginError(f'PCN: {pcn} is not present in reference file. Verify pcn.json data.')
        _url = self.driver.current_url
        if self.single_pcn:
            warn(f'This account only has access to one PCN.', loglevel=2)
            return
        if not 'MENUCUSTOMER.ASPX' in _url.upper() and self.first_login:
            self.debug_logger.debug(f'Single PCN account detected.')
            self.single_pcn = True
            return
        if not self.single_pcn and not self.first_login:
            self.driver.get(f'{self.url_comb}/Modules/SystemAdministration/MenuSystem/MenuCustomer.aspx')
        if self._pcn_link_locate(f'//img[@alt="{_pcn_name}"]'):
            return
        if self._pcn_link_locate(f'//*[contains(text(), "{_pcn_name}")]'):
            return
        raise LoginError(self.environment, self.db, _pcn_name, f'Unable to locate PCN. Verify you have access.')


    def _pcn_link_locate(self, xpath):
        try:
            self.driver.find_element(By.XPATH, xpath).click()
            return True
        except NoSuchElementException:
            return False


    def token_get(self) -> str:
        """Sets self.url_comb for current token value.

        Returns:
            str: Plex URL combination containing token.
        """
        url = self.driver.current_url
        url_split = url.split('/')
        url_proto = url_split[0]
        url_domain = url_split[2]
        self.url_token = url_split[3]
        self.url_comb = f'{url_proto}//{url_domain}/{self.url_token}'
        return self.url_comb

class ClassicPlexElement(PlexElement):
    def __init__(self, webelement, parent):
        super().__init__(webelement, parent)

    def sync_picker(self, text_content, clear=False, date=False):
        """TODO - Create this function."""


from pmc_automation_tools.api.common import DataSourceInput, DataSourceResponse, DataSource
from pmc_automation_tools.common.exceptions import ClassicConnectionError

import requests
from requests.auth import HTTPBasicAuth

from zeep import Client
from zeep.transports import Transport
from zeep.helpers import serialize_object

SOAP_TEST = 'https://testapi.plexonline.com/Datasource/service.asmx'
SOAP_PROD = 'https://api.plexonline.com/Datasource/service.asmx'
class ClassicDataSourceInput(DataSourceInput):
    """Input object that stores the attributes for building the proper request format."""
    def __init__(self, data_source_key: int, *args, delimeter='|', **kwargs):
        self._delimeter = delimeter
        super().__init__(data_source_key, *args, type='classic', **kwargs)
        self.__api_id__ = int(self.__api_id__)

    def _update_input_parameters(self):
        self._parameter_names = self._delimeter.join([k for k, v in vars(self).items() if not k.startswith('_')])
        self._parameter_values = self._delimeter.join([str(v) for k, v in vars(self).items() if not k.startswith('_')])


class ClassicDataSource(DataSource):
    def __init__(self, auth: HTTPBasicAuth|str, wsdl,
                 *args,
                 test_db: bool = True,
                 pcn_config_file: str='resources/pcn_config.json',
                 **kwargs):
        """Data Source object for Classic SOAP web service

        Args:
            auth (HTTPBasicAuth | str): Authentication for web service account
            wsdl (str): path to the SOAP wsdl file
            test_db (bool, optional): Connect to the test api URL. Defaults to True.
            pcn_config_file (str, optional): path to the web service credential file. Defaults to 'resources/pcn_config.json'.
        """
        super().__init__(*args, auth=auth, test_db=test_db, pcn_config_file=pcn_config_file, type='classic', **kwargs)
        self._wsdl = wsdl


    def call_data_source(self, query:ClassicDataSourceInput) -> 'ClassicDataSourceResponse':
        """Triggers the data source request.

        Args:
            query (ClassicDataSourceInput): object containing the input details

        Returns:
            ClassicDataSourceResponse: ClassicDataSourceResponse object
        """
        session = requests.Session()
        session.auth = self._auth
        client = Client(wsdl=self._wsdl, transport=Transport(session=session))
        self._connection_address = client.wsdl.services['Service'].ports['ServiceSoap'].binding_options['address']
        if self._test_db and self._connection_address != SOAP_TEST:
            raise
        response = client.service.ExecuteDataSourcePost(dataSourceKey=query.__api_id__, parameterNames=query._parameter_names, parameterValues=query._parameter_values, delimeter=query._delimeter)
        _response = serialize_object(response, dict)
        return ClassicDataSourceResponse(query.__api_id__, **_response)


class ClassicDataSourceResponse(DataSourceResponse):
    def __init__(self, data_source_key, **kwargs):
        super().__init__(data_source_key, **kwargs)
        if self.Error:
            raise ClassicConnectionError(self.Message,
                                         data_source_key=self.DataSourceKey,
                                         instance=self.InstanceNo,
                                         status=self.StatusNo,
                                         error_no=self.ErrorNo)
        self._result_set = kwargs.get('ResultSets')
        if self._result_set:
            self._row_count = self._result_set['ResultSet'][0]['RowCount']
            self._result_set = self._result_set['ResultSet'][0]['Rows']['Row']
            self._format_response()
    
    
    def _format_response(self):
        self._transformed_data = []
        if hasattr(self, '_result_set'):
            for row in self._result_set:
                row_data = {}
                columns = row['Columns']['Column']
                for column in columns:
                    name = column['Name']
                    value = column['Value']
                    row_data[name] = value
                self._transformed_data.append(row_data)
        return self._transformed_data


class Error(Exception):
    pass
class PmcAutomationToolsError(Error):
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        for key, value in kwargs.items():
            setattr(self, key, value)
class PlexAutomateError(PmcAutomationToolsError):
    """A base class for handling exceptions in this project"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        for key, value in kwargs.items():
            setattr(self, key, value)
class NoRecordError(PlexAutomateError):
    """Thrown when no records exist in a picker selection."""
class ActionError(PlexAutomateError):
    """Thrown if there is an error on clicking an action bar item."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.expression = kwargs.get('expression')
        self.message = kwargs.get('message')
class LoginError(PlexAutomateError):
    """Thrown if the expected login screens are not found."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.environtment = kwargs.get('environment')
        self.db = kwargs.get('db')
        self.pcn = kwargs.get('pcn')
        self.message = kwargs.get('message')
class UpdateError(PlexAutomateError):
    """Thrown when a banner prevents an update from occurring."""
    def __init__(self, *args, **kwargs):
        super().__init__(*args)
        self.clean_message = args[0].replace('×', '').replace('\n', '').strip()

class PlexApiError(PmcAutomationToolsError):
    """A base class for handling exceptions for API calls"""
class PlexResponseError(PlexApiError):...
class DataSourceError(PlexApiError):...
class ApiError(DataSourceError):...
class ClassicConnectionError(DataSourceError):...

class UXResponseError(PlexResponseError):
    def __init__(self, error_dict):
        self.code = error_dict.get('code')
        self.message = error_dict.get('message')

    def __str__(self):
        return f"{self.code}: {self.message}"
    def __repr__(self):
        return f"<ErrorDetail(code={self.code}, message={self.message})>"
class UXResponseErrorLog(PlexResponseError):
    def __init__(self, error_dicts):
        self.errors = [UXResponseError(err) for err in error_dicts]
    
    def __getitem__(self, index):
        return self.errors[index]

    def __len__(self):
        return len(self.errors)

    def __str__(self):
        return "\n".join(str(error) for error in self.errors)

    def __repr__(self):
        return f"<UXResponseErrorLog with {len(self)} errors>"

    def print_all_errors(self):
        for error in self.errors:
            print(error)

    def filter_by_code(self, code):
        return [error for error in self.errors if error.code == code]
        


from typing import Literal
from pmc_automation_tools.driver.common import (
    PlexDriver,
    PlexElement,
    VISIBLE,
    INVISIBLE,
    CLICKABLE,
    EXISTS,
    SIGNON_URL_PARTS
    )
from selenium.common.exceptions import (
    TimeoutException,
    StaleElementReferenceException,
    NoSuchElementException
    )
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from pmc_automation_tools.common.exceptions import (
    UpdateError,
    NoRecordError,
    LoginError
    )
import time
BANNER_SUCCESS = 1
BANNER_WARNING = 2
BANNER_ERROR = 3
BANNER_CLASSES = {
    'plex-banner-success': BANNER_SUCCESS,
    'plex-banner-error': BANNER_WARNING,
    'plex-banner-warning': BANNER_ERROR
}
BANNER_SELECTOR = (By.CLASS_NAME, 'plex-banner')
PLEX_GEARS_SELECTOR = (By.XPATH, '//i[@class="plex-waiting-spinner"]')
UX_INVALID_PCN_MESSAGE = '__MESSAGE=YOU+WERE+REDIRECTED+TO+YOUR+LANDING+COMPANY'


class UXDriver(PlexDriver):
    """
    UX Unique requirements
    no pcn_file required
    url parts different
    doesn't need csv_to_json function
    
    wait_for_banner
    wait_for_gears
    _banner_handler

    login uniqueness

    """
    def __init__(self, driver_type: Literal['edge', 'chrome'], *args, **kwargs):
        super().__init__(environment='ux', *args, driver_type=driver_type, **kwargs)
        for k, v in kwargs.items():
            setattr(self, k, v)
    def wait_for_element(self, selector, driver=None, timeout=15, type=VISIBLE, ignore_exception=False):
        return super().wait_for_element(selector, driver=driver, timeout=timeout, type=type, ignore_exception=ignore_exception, element_class=UXPlexElement)
    def wait_for_banner(self) -> None:
        try:
            loop = 0
            while loop <= 10:
                banner = self.wait_for_element(BANNER_SELECTOR)
                banner_class = banner.get_attribute('class')
                banner_type = next((BANNER_CLASSES[c] for c in BANNER_CLASSES if c in banner_class), None)
                if banner_type:
                    self._banner_handler(banner_type, banner)
                    break
                time.sleep(1)
                loop += 1
            else:
                raise UpdateError(f'Unexpected banner type detected. Found {banner_class}. Expected one of {list(BANNER_CLASSES.keys())}')
        except (TimeoutException, NoSuchElementException, StaleElementReferenceException):
            raise UpdateError('No banner detected.')


    def _banner_handler(self, banner_type, banner):
        if banner_type == BANNER_SUCCESS:
            return 
        else:
            banner_text = banner.get_property('textContent')
            raise UpdateError(banner_text)
    

    def wait_for_gears(self, loading_timeout=10):
        super().wait_for_gears(PLEX_GEARS_SELECTOR, loading_timeout)

    def click_button(self, button_text, driver=None):
        """
        Clicks a standard button when given the text.
        
        Mainly used for clicking the Ok and Apply buttons.

        Can be used for Search buttons as well.
        
        Parameters:
        
        - name: Text of the button to click
        - driver: webdriver root to use if different than default

        Usage:

            If you don't provide the root driver, then the main page's Ok button will be clicked and not the popup window's button.
            ::

                popup_window = driver.find_element(By.ID, 'popupID')
                click_button('Ok', driver=popup_window)

            Alternatively:
            ::

                pa = UXDriver()
                popup_window = pa.wait_for_element((By.ID, 'popupID'))
                popup_window.click_button('Ok')
                
        """
        driver = driver or self.driver
        buttons = driver.find_elements(By.TAG_NAME, 'button')
        for b in buttons:
            if b.get_property('textContent') == button_text:
                self.debug_logger.debug(f'Button found with matching text: {button_text}')
                b.click()
                break
            
    def click_action_bar_item(self, item, sub_item=None):
        """
        Clicks on an action bar item.

        Parameters:
        
        - item: Text for the item to click
        - sub_item: Text for the item if it is within a dropdown from clicking the item first.
        """
        action_bar = self.wait_for_element((By.CLASS_NAME, 'plex-actions'))

        # Check for the "More" link and determine if it's visible
        try:
            more_box = action_bar.find_element(By.LINK_TEXT, "More")
            style = more_box.find_element(By.XPATH, 'ancestor::li').get_dom_attribute('style')
            more_visible = 'none' not in style
            self.debug_logger.debug(f"'More' link {'found and visible' if more_visible else 'found but not visible'}.")
        except NoSuchElementException:
            self.debug_logger.debug('No element found for "More" link.')
            more_visible = False
        # Click on "More" if visible and adjust the action bar
        if more_visible:
            self.debug_logger.debug('Clicking "More" button.')
            more_box.click()
            self.wait_for_element((By.CLASS_NAME, "plex-subactions.open"))
            action_bar = self.wait_for_element((By.CLASS_NAME, 'plex-actions-more'))

        # Handle sub_item or main item click
        if sub_item:
            self.self.debug_logger.debug("Clicking sub-item.")
            self._click_sub_item(action_bar, item, sub_item)
        else:
            self.self.debug_logger.debug("Clicking main item.")
            action_item = self.wait_for_element((By.LINK_TEXT, item), type=CLICKABLE)
            action_item.click()

    def _click_sub_item(self, action_bar, item, sub_item):
        """Helper function to click on a sub-item."""
        action_items = action_bar.find_elements(By.CLASS_NAME, "plex-actions-has-more")
        for a in action_items:
            span_texts = a.find_elements(By.TAG_NAME, 'span')
            for s in span_texts:
                if s.get_property('textContent') == item:
                    s.find_element(By.XPATH, "ancestor::a").click()
                    break
        action_bar.find_element(By.LINK_TEXT, sub_item).click()


    def login(self, username, password, company_code, pcn, test_db=True, headless=False):
        self._set_login_vars()
        super().login(username, password, company_code, pcn, test_db, headless)
        self._login_validate()
        self.pcn_switch(self.pcn)
        self.token = self.token_get()
        self.first_login = False
        return (self.driver, self.url_comb, self.token)
    
    
    def _set_login_vars(self):
        self.plex_main = 'cloud.plex.com'
        self.plex_prod = ''
        self.plex_test = 'test.'
        self.sso = '/sso'
        super()._set_login_vars()


    def token_get(self):
        url = self.driver.current_url
        url_split = url.split('/')
        url_proto = url_split[0]
        url_domain = url_split[2]
        self.url_comb = f'{url_proto}//{url_domain}'
        self.url_token = url.split('?')[1]
        if '&' in self.url_token:
            self.url_token = [x for x in self.url_token.split('&') if 'asid' in x][0]
        return self.url_token
    
    
    def _pcn_switch(self, pcn=None):
        if not pcn:
            pcn = self.pcn
        if self.first_login:
            self.first_login = False
            return
        self.url_token = self.token_get()
        self.driver.get(f'{self.url_comb}/SignOn/Customer/{pcn}?{self.url_token}')
        if UX_INVALID_PCN_MESSAGE in self.driver.current_url.upper():
            raise LoginError(self.environment, self.db, pcn, f'Unable to login to PCN. Verify you have access.')
    
    
    def _login_validate(self):
        url = self.driver.current_url
        if not any(url_part in url.upper() for url_part in SIGNON_URL_PARTS):
            raise LoginError(self.environment, self.db, self.pcn_name, 'Login page not detected. Please validate login credentials and try again.')
        


class UXPlexElement(PlexElement):
    """
    Unique functions
        sync_picker
            Will be highly customized to how UX pickers work.
            Classic pickers would need spearate logic
    element class functions
        sync_textbox
        sync_checkbox
        save_element_image
    driver class functions
        wait_for_gears UXDriver inheritance
        wait_for_element PlexElement inheritance
    """
    def __init__(self, webelement, parent):
        super().__init__(webelement, parent)


    def sync_picker(self, text_content, clear=False, date=False):
            """
            Sync the picker element to the provided value.

            parameters

                : text_content - Desired value for the picker
                : clear - Clear out the picker if providing a blank text_content
                : date - If the picker is a date picker. This should be detected automatically, but can be forced if behavior is unexpected.
            """
            multi = False
            matching = False
            picker_type = self.get_attribute('class')
            if picker_type == 'input-sm':
                date = True
            if not text_content and not clear:
                return
            if self.tag_name == 'select':
                self.debug_logger.debug(f'Picker type is selection list.')
                _select = Select(self)
                _current_selection = _select.first_selected_option.text
                if _current_selection == text_content:
                    self.debug_logger.debug(f'Picker selection: {_current_selection} matches {text_content}')
                    return
                for o in _select.options:
                    if o.text == text_content:
                        matching = True
                if matching:
                    self.debug_logger.info(f'Matching option found. Picking {text_content}')
                    _select.select_by_visible_text(text_content)
                    self.send_keys(Keys.TAB)
                else:
                    self.debug_logger.info(f'No matching selection available for {text_content}')
                    raise NoRecordError(f'No matching selection available for {text_content}')
                return
            else:
                try:
                    # We would then need to locate if a value is already input and check the title attribute
                    self.debug_logger.debug(f'Trying to locate an existing selected item.')
                    selected_element = self.wait_for_element((By.XPATH, "preceding-sibling::div[@class='plex-picker-selected-items']"), driver=self, timeout=1)
                    if selected_element:
                        self.debug_logger.debug(f'Selected item detected')
                        selected_item = self.wait_for_element((By.CLASS_NAME, "plex-picker-item-text"), driver=selected_element)
                        current_text = selected_item.get_property('textContent')
                        if current_text != text_content:
                            self.debug_logger.info(f'Current text: {current_text} does not match provided text: {text_content}')
                            self.send_keys(Keys.BACKSPACE) # Backspace will convert the picker item to normal text.
                            self.clear()
                        else:
                            self.debug_logger.debug(f'Current text: {current_text} matches provided text: {text_content}.')
                            matching = True
                            return
                except (NoSuchElementException, TimeoutException):
                    self.debug_logger.debug(f'No initial selected item.')
                finally:
                    if matching:
                        self.debug_logger.debug(f'Existing value matches.')
                        return
                    self.send_keys(text_content)
                    self.send_keys(Keys.TAB)
                    try:
                        if date:
                            self.debug_logger.debug(f'Picker is a date filter.')
                            self.wait_for_element((By.XPATH, "preceding-sibling::div[@class='plex-picker-item']"), driver=self, timeout=5)
                            self.debug_logger.info(f'Date picker has been filled in with {text_content}')
                        else:
                            self.wait_for_element((By.XPATH, "preceding-sibling::div[@class='plex-picker-selected-items']"), driver=self, timeout=5)
                            self.debug_logger.info(f'Normal picker has been filled in with {text_content}')
                    except (TimeoutException, NoSuchElementException) as e:
                        try:
                            self.debug_logger.debug(f'No auto filled item, checking for a popup window.')
                            popup = self.wait_for_element((By.CLASS_NAME, 'modal-dialog.plex-picker'), timeout=3)
                            if 'plex-picker-multi' in popup.get_attribute('class'):
                                self.debug_logger.debug(f'Picker is a multi-picker')
                                multi = True
                            self.wait_for_gears()
                            items = popup.find_elements(By.CLASS_NAME, 'plex-grid-row')
                            if not items:
                                result_text = popup.find_element(By.TAG_NAME, 'h4').get_property('textContent')
                                if 'No records' in result_text:
                                    self.debug_logger.info(f'No records found for {text_content}')
                                    footer = popup.find_element(By.CLASS_NAME, 'modal-footer')
                                    _cancel = footer.find_element(By.LINK_TEXT, 'Cancel')
                                    _cancel.click()
                                    raise NoRecordError(f'No records found for {text_content}')
                            for i in items:
                                if i.text != text_content:
                                    continue
                                self.debug_logger.info(f'Found matching item with text {i.text}.')
                                i.click()
                                if multi:
                                    self.debug_logger.info(f'Multi-picker, clicking ok on the popup window.')
                                    self.click_button('Ok', driver=popup)
                                    self.debug_logger.info(f'Multi-picker, clicked ok on the popup window.')
                                    
                        except (TimeoutException, NoSuchElementException) as e:
                            self.debug_logger.info(f'No matching elements found for {text_content}')



import urllib
import zipfile
import winreg
import re
from warnings import warn

from pmc_automation_tools.common.get_file_properties import getFileProperties
import os
import time
import json
from abc import ABC, abstractmethod
from typing import Literal
from pathlib import Path

from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.by import By
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import (
    TimeoutException,
    StaleElementReferenceException,
    NoSuchElementException
    )
from selenium.webdriver.chrome.service import Service
from selenium.webdriver.remote.webelement import WebElement
from pmc_automation_tools.common.exceptions import (
    UpdateError,
    LoginError,
)
from pmc_automation_tools.common.utils import (
    debug_logger,
    create_batch_folder,
    )
VALID_ENVIRONMENTS = {'ux', 'classic'}


BANNER_SUCCESS = 1
BANNER_WARNING = 2
BANNER_ERROR = 3
BANNER_CLASSES = {
    'plex-banner-success': BANNER_SUCCESS,
    'plex-banner-error': BANNER_WARNING,
    'plex-banner-warning': BANNER_ERROR
}
BANNER_SELECTOR = (By.CLASS_NAME, 'plex-banner')

SIGNON_URL_PARTS = {'/LAUNCHPAGE', '/MENUCUSTOMER.ASPX', '/MENU.ASPX'}

VISIBLE = 10
INVISIBLE = 20
CLICKABLE = 30
EXISTS = 0
_wait_until = {
    VISIBLE : EC.presence_of_element_located,
    INVISIBLE : EC.invisibility_of_element_located,
    CLICKABLE : EC.element_to_be_clickable,
    EXISTS : None
}


class PlexDriver(ABC):
    def __init__(self, environment: Literal['ux', 'classic'], *args, driver_type: Literal['edge', 'chrome']='edge', **kwargs):
        for k, v in kwargs.items():
            setattr(self, k, v)
        self.driver_type = driver_type
        self.pcn_file_path = kwargs.get('pcn_file_path', Path('resources/pcn.json'))
        self.debug = kwargs.get('debug', False)
        self.debug_level = kwargs.get('debug_level', 0)
        self.debug_logger = debug_logger(self.debug_level)
        self.environment = environment.lower()
        self.driver_override = kwargs.get('driver_override')
        self.single_pcn = False
        
        self._set_login_vars()
        self._path_init()
        self._read_driver_version()
        self.debug_logger.debug('finished initializing.')


    def _set_login_vars(self):
        self.plex_log_id = 'username'
        self.plex_log_pass = 'password'
        self.plex_log_comp = 'companyCode'
        self.plex_login = 'loginButton'


    def _path_init(self):
        self.resource_path = 'resources'
        if not os.path.exists(self.resource_path):
            os.mkdir(self.resource_path)
        self.download_dir = 'downloads'
        if not os.path.exists(self.download_dir):
            os.mkdir(self.download_dir)
        self.latest_driver_version_file = os.path.join(self.resource_path, 'driver_version.txt')


    def wait_for_element(self, selector:tuple[str, str], driver=None, timeout:int=15, type=VISIBLE, ignore_exception: bool=False, element_class:object=None) -> 'PlexElement':
        """Wait until an element meets specified criteria.

        Wrapper for Selenium WebDriverWait function for common Plex usage.

        Args:
            selector (tuple[str, str]): Selenium style element selector e.g. (By.NAME, 'ElementName')
            driver (WebDriver|PlexElement, optional): root WebDriver to use for locating the element. Defaults to None.
            timeout (int, optional): Time to wait until timeout is triggered. Defaults to 15.
            type (str, optional): Type of wait to be performed. Defaults to VISIBLE.
            ignore_exception (bool, optional): Don't raise an exception if the timeout is reached. Defaults to False.
            element_class (object, optional): class of the WebElement to return. Defaults to None.

        Raises:
            exception: re-raises exception from failing to meet the expected condition

        Returns:
            PlexElement: PlexElement based on search criteria
        """
        try:
            driver = driver or self.driver
            element_condition = _wait_until.get(type)
            if element_condition:
                WebDriverWait(driver, timeout).until(element_condition(selector))
            element_class = element_class or PlexElement
            return element_class(driver.find_element(*selector), self)

        except (TimeoutException, StaleElementReferenceException, NoSuchElementException):
            if ignore_exception:
                return None
            raise
    
    def search_for_element(self, selector, match_value, driver=None, ignore_exception=False):
        try:
            driver = driver or self.driver
            _el = driver.find_elements(*selector)
            for e in _el:
                val = e.get_attribute('value')
                tex = e.get_attribute('textContent')
                if match_value == val or match_value == tex:
                    return e
            raise NoSuchElementException('No element could be found with the provided selector and match value.')
        except (TimeoutException, StaleElementReferenceException, NoSuchElementException):
            if ignore_exception:
                return None
            raise

    def wait_for_banner(self) -> None:
        try:
            loop = 0
            while loop <= 10:
                banner = self.wait_for_element(BANNER_SELECTOR)
                banner_class = banner.get_attribute('class')
                banner_type = next((BANNER_CLASSES[c] for c in BANNER_CLASSES if c in banner_class), None)
                if banner_type:
                    self._banner_handler(banner_type, banner)
                    break
                time.sleep(1)
                loop += 1
            else:
                raise UpdateError(f'Unexpected banner type detected. Found {banner_class}. Expected one of {list(BANNER_CLASSES.keys())}')
        except (TimeoutException, NoSuchElementException, StaleElementReferenceException):
            raise UpdateError('No banner detected.')


    def _banner_handler(self, banner_type, banner):
        if banner_type == BANNER_SUCCESS:
            return
        else:
            banner_text = banner.get_property('textContent')
            raise UpdateError(banner_text)

    def wait_for_gears(self, selector, loading_timeout=10):
        """Wait for the spinning gears image to appear and disappear

        This should be called after searching or updating a screen.

        Essentially any time you are clicking a button which would cause the page to load.

        The gears sometimes dissappear quicker than can be detected. 
            If the gears are not detected at the begining, the end timeout is shorter.

        Arg:
            selector (tuple[str, str]): Selenium style element selector e.g. (By.NAME, 'ElementName')
            loading_timeout (int, optional): Time to wait until the gears disappear after being detected. Defaults to 10.
        """
        gears_visible = False
        gears_visible = self.wait_for_element(selector, type=VISIBLE, timeout=1, ignore_exception=True)
        timeout = loading_timeout if gears_visible else 1
        self.debug_logger.debug(f'Timeout for invisible is {timeout}.')
        self.wait_for_element(selector, type=INVISIBLE, timeout=timeout, ignore_exception=True)

    def login(self, username, password, company_code, pcn, test_db=True, headless=False):
        """Log in to Plex

        Args:
            username (str): Plex username
            password (str): Plex password
            company_code (str): Plex company code
            pcn (str): PCN number
            test_db (bool, optional): Log in to the test database. Defaults to True.
            headless (bool, optional): Run in headless mode. Defaults to False.
        """
        self.test_db = test_db
        self.batch_folder = create_batch_folder(test=self.test_db)
        self.pcn = pcn
        self.headless = headless
        if hasattr(self, 'pcn_dict'):
            self.pcn_name = self.pcn_dict[self.pcn]
        else:
            self.pcn_name = self.pcn
        self.driver = self._driver_setup(self.driver_type)

        db = self.plex_test if self.test_db else self.plex_prod
        self.driver.get(f'https://{db}{self.plex_main}{self.sso}')
        # Test for new login screen
        try:
            self.wait_for_element((By.XPATH, '//img[@alt="Rockwell Automation"]'), timeout=4)
            self.debug_logger.debug(f'New Rockwell IAM login screen detected.')
            rockwell = True
        except (NoSuchElementException, TimeoutException):
            self.wait_for_element((By.XPATH, '//img[@alt="Plex"]'))
            rockwell = False
        if rockwell:
            id_box = self.wait_for_element((By.NAME, self.plex_log_id), type=CLICKABLE)
            id_box.send_keys(username)
            id_box.send_keys(Keys.TAB)
            company_box = self.wait_for_element((By.NAME, self.plex_log_comp), type=CLICKABLE)
            company_text = company_box.get_attribute('value')
            if company_text != company_code:
                self.debug_logger.info(f'Auto-populated company code: {company_text} does not match provided login credentials: {company_code}.')
                company_box.click()
                company_box.clear()
                company_box.send_keys(company_code)
                company_box.send_keys(Keys.TAB)
            pass_box = self.wait_for_element((By.NAME, self.plex_log_pass), type=CLICKABLE)
            pass_box.send_keys(password)
        else:
            self.debug_logger.debug(f'Plex IAM login screen detected.')
            id_box = self.driver.find_element(By.NAME, self.plex_log_id)
            pass_box = self.driver.find_element(By.NAME, self.plex_log_pass)
            company_code_box = self.driver.find_element(By.NAME, self.plex_log_comp)
            company_code_box.send_keys(company_code)
            company_code_box.send_keys(Keys.TAB)
            id_box.send_keys(username)
            id_box.send_keys(Keys.TAB)
            pass_box.send_keys(password)
        login_button = self.wait_for_element((By.ID, self.plex_login), type=CLICKABLE)
        login_button.click()
        self.first_login = True

    def _driver_setup(self, type):
        if type == 'edge':
            self._edge_check()
            self._download_edge_driver(self.full_browser_version)  # Adjust this to download the correct Edge driver
            return self._edge_setup()
        if type == 'chrome':
            self._chrome_check()
            self._download_chrome_driver(self.driver_override)
            return self._chrome_setup()


    def _read_driver_version(self):
        if os.path.exists(self.latest_driver_version_file):
            with open(self.latest_driver_version_file, 'r') as f:
                self.latest_downloaded_driver_version = f.read()
                self.debug_logger.debug(f'Latest downloaded chromedriver version: {self.latest_downloaded_driver_version}')
        else:
            self.debug_logger.debug(f'Latest downloaded chromedriver file not detected. Setting version to None.')
            self.latest_downloaded_driver_version = None


    def _save_driver_version(self, version):
        self.debug_logger.debug(f'Saving latest downloaded driver version: {version}')
        with open(self.latest_driver_version_file, 'w+') as f:
            f.write(version)




    def _download_edge_driver(self, version:str=None):
        """Downloads the EdgeDriver that will allow Selenium to function.

        Args:
            version (str, optional): edgedriver version if different than latest. Defaults to None.
        """
        text_path = os.path.join(self.resource_path, 'edgedriver.txt')
        zip_path = os.path.join(self.resource_path, 'edgedriver.zip')
        edgedriver_url = None
        if version:
            self.full_browser_version = version
        else:
            url = 'https://msedgedriver.azureedge.net/LATEST_STABLE'

            urllib.request.urlretrieve(url, text_path)
            with open(text_path, 'r', encoding='utf-16') as f:
                latest_edgedriver_version = f.read().strip()
            self.full_browser_version = latest_edgedriver_version
        edgedriver_url = f'https://msedgedriver.azureedge.net/{self.full_browser_version}/edgedriver_win64.zip'
        
        if self.latest_downloaded_driver_version == edgedriver_url:
            self.debug_logger.debug(f'Latest downloaded EdgeDriver version matches the latest online. Skipping download.')
            return
        else:
            self._save_driver_version(edgedriver_url)

        self.debug_logger.debug(f'Downloading EdgeDriver version from: {edgedriver_url}')
        urllib.request.urlretrieve(edgedriver_url, zip_path)
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            self.debug_logger.debug(f'Extracting EdgeDriver to: {self.resource_path}')
            zip_ref.extractall(self.resource_path)
            
    def _download_chrome_driver(self, version:str=None):
        """Downloads the ChromeDriver that will allow Selenium to function.

        Args:
            version (str, optional): chromedriver version if different than latest. Defaults to None.
        """
        text_path =os.path.join(self.resource_path, 'chromedriver.txt')
        zip_path =os.path.join(self.resource_path, 'chromedriver.zip')
        chromedriver_url = None
        url = 'https://googlechromelabs.github.io/chrome-for-testing/latest-versions-per-milestone-with-downloads.json'
        urllib.request.urlretrieve(url, text_path)
        with open(text_path, 'r') as f:
            chrome_releases = json.load(f)
        latest_chromedriver_version = chrome_releases['milestones'][self.browser_version]['downloads']['chromedriver']#['platform']
        
        for x in latest_chromedriver_version:
            if x['platform'] == 'win64':
                chromedriver_url = x['url']
        if chromedriver_url:
            url = chromedriver_url
            if self.latest_downloaded_chromedriver_version == url:
                self.debug_logger.debug(f'Latest downloaded chromedriver version matches with lasted on web. Skipping download.')
                return
            else:
                self._save_driver_version(url)
        else:
            if version:
                self.full_browser_version = version
            url = f'https://edgedl.me.gvt1.com/edgedl/chrome/chrome-for-testing/{self.full_browser_version}/win64/chromedriver-win64.zip'
        self.debug_logger.debug(f'Downloading chromedriver version at: {url}')
        urllib.request.urlretrieve(url, zip_path)
        with zipfile.ZipFile(zip_path, 'r') as zip_ref:
            self.debug_logger.debug(f'Exctracting chromedriver to: {self.resource_path}')
            zip_ref.extractall(self.resource_path)

    def _chrome_check(self):
        chrome_key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "ChromeHTML\\shell\\open\\command", 0, winreg.KEY_READ)
        command = winreg.QueryValueEx(chrome_key, "")[0]
        winreg.CloseKey(chrome_key)
        chrome_browser = re.search("\"(.*?)\"", command)
        if chrome_browser:
            chrome_browser = chrome_browser.group(1)
            print(f'Chrome browser install location: {chrome_browser}')
        cb_dictionary = getFileProperties(chrome_browser) # returns whole string of version (ie. 76.0.111)
        self.browser_version = cb_dictionary['FileVersion'].split('.')[0] # substring version to capabable version (ie. 77 / 76)
        self.full_browser_version = cb_dictionary['FileVersion']
        print(f'Chrome base version: {self.browser_version} | Full Chrome version: {self.full_browser_version}')
    
    def _edge_check(self):
        edge_key = winreg.OpenKey(winreg.HKEY_CLASSES_ROOT, "MSEdgeHTM\\shell\\open\\command", 0, winreg.KEY_READ)
        command = winreg.QueryValueEx(edge_key, "")[0]
        winreg.CloseKey(edge_key)
        edge_browser = re.search("\"(.*?)\"", command)
        if edge_browser:
            edge_browser = edge_browser.group(1)
            print(f'Edge browser install location: {edge_browser}')
        edge_dictionary = getFileProperties(edge_browser)  # returns the full version string (e.g., 91.0.864.67)
        self.browser_version = edge_dictionary['FileVersion'].split('.')[0]  # Extract the major version (e.g., 91)
        self.full_browser_version = edge_dictionary['FileVersion']  # Full version (e.g., 91.0.864.67)

        print(f'Edge base version: {self.browser_version} | Full Edge version: {self.full_browser_version}')

    def _edge_setup(self):
        executable_path = os.path.join(self.resource_path, 'msedgedriver.exe')
        edge_options = EdgeOptions()
        edge_options.use_chromium = True
        edge_options.add_argument("--log-level=3")
        if self.headless:
            self.debug_logger.debug(f'Running Edge in headless mode.')
            edge_options.add_argument("--headless")
        edge_options.add_experimental_option("prefs", {
            "download.default_directory": f"{self.download_dir}",
            "download.prompt_for_download": False,
        })

        service = Service(executable_path=executable_path)
        return webdriver.Edge(service=service, options=edge_options)
    

    def _chrome_setup(self):
        executable_path = os.path.join(self.resource_path, 'chromedriver-win64', 'chromedriver.exe')
        os.environ['webdriver.chrome.driver'] = executable_path
        chrome_options = Options()
        chrome_options.add_argument("--log-level=3")
        if self.headless:
            self.debug_logger.debug(f'Running chrome in headless mode.')
            chrome_options.add_argument("--headless")
        chrome_options.add_experimental_option("prefs", {
            "download.default_directory": f"{self.download_dir}",
            "download.prompt_for_download": False,
            })
        service = Service(executable_path=executable_path)
        return webdriver.Chrome(service=service, options=chrome_options)


    @abstractmethod
    def token_get(self):...

    @abstractmethod
    def _pcn_switch(self):...

    
    def pcn_switch(self, pcn):
        pcn = str(pcn)
        self.debug_logger.debug(f'Switching to PCN: {pcn}.')
        self._pcn_switch(pcn)
        return self.token_get()    
    switch_pcn = pcn_switch

    @abstractmethod
    def click_button(self):...
    
class PlexElement(WebElement):
    """
    Subclass of Selenium WebElement with specialized functions for Plex elements.
    """
    def __init__(self, webelement, parent):
        super().__init__(webelement._parent, webelement._id)
        self.debug = parent.debug
        self.debug_level = parent.debug_level
        self.debug_logger = debug_logger(self.debug_level)
        self.batch_folder = parent.batch_folder
        self.test_db = parent.test_db
        self.driver = webelement._parent
        self.wait_for_element = parent.wait_for_element
        self.click_button = parent.click_button
        self.wait_for_gears = parent.wait_for_gears
        self.search_for_element = parent.search_for_element


    def screenshot(self):
        """
        Save a screenshot of the element. Useful to debug if there are any issues locating the element properly.
        """
        element_id = self.id[-8:]
        session = self.parent.session_id[-5:]
        name = self.accessible_name or 'No_Name'
        if not hasattr(self, 'batch_folder'):
            self.batch_folder = create_batch_folder(test=self.test_db)
        if not hasattr(self, 'screenshot_folder'):
            self.screenshot_folder = os.path.join(self.batch_folder, 'screenshots')
            os.makedirs(self.screenshot_folder, exist_ok=True)
        filename = os.path.join(self.screenshot_folder, f"{session}_{element_id}_{name}_screenshot.png")
        super().screenshot(filename)

    
    def sync_checkbox(self, bool_state:bool|int):
        """Sync a checkbox to the provided checked state

        Args:
            bool_state (bool | int): Checkbox state to make the element.
        """
        if not type(bool_state) == bool:
            bool_state = bool(int(bool_state))
        check_state = self.get_property('checked')
        if not check_state == bool_state:
            self.debug_logger.info(f'{self.get_property("name")} - Checkbox state: {check_state}. Clicking to make it {bool_state}.')
            self.click()
        else:
            self.debug_logger.debug(f'{self.get_property("name")} - Checkbox state: {check_state} matches provided state: {bool_state}.')


    def sync_textbox(self, text_content:str, clear:bool=False):
        """Sync a textbox with the provided value

        Args:
            text_content (str): Desired value for the text box
            clear (bool, optional): Clear out the text box if sending an empty string. Defaults to False.
        """
        if not text_content and not clear:
            return
        text = self.get_property('value')
        if not text == text_content:
            text_content = text_content.replace('\t', ' ') # The input will break if sending tab characters. This should only happen when a copy/paste from word/excel was done on the original field text.
            self.debug_logger.info(f'{self.get_property("name")} - Current text: {text}. Replacing with provided text: {text_content}')
            self.clear()
            self.send_keys(text_content)
            self.send_keys(Keys.TAB)
        else:
            self.debug_logger.debug(f'{self.get_property("name")} - Current text: {text}. Matches provided text: {text_content}')

    @abstractmethod
    def sync_picker(self):...


# as per https://stackoverflow.com/questions/580924/python-windows-file-version-attribute

import win32api

#==============================================================================
def getFileProperties(fname):
#==============================================================================
    """
    Read all properties of the given file return them as a dictionary.
    """
    propNames = ('Comments', 'InternalName', 'ProductName',
        'CompanyName', 'LegalCopyright', 'ProductVersion',
        'FileDescription', 'LegalTrademarks', 'PrivateBuild',
        'FileVersion', 'OriginalFilename', 'SpecialBuild')

    props = {'FixedFileInfo': None, 'StringFileInfo': None, 'FileVersion': None}

    try:
        # backslash as parm returns dictionary of numeric info corresponding to VS_FIXEDFILEINFO struc
        fixedInfo = win32api.GetFileVersionInfo(fname, '\\')
        props['FixedFileInfo'] = fixedInfo
        props['FileVersion'] = "%d.%d.%d.%d" % (fixedInfo['FileVersionMS'] / 65536,
                fixedInfo['FileVersionMS'] % 65536, fixedInfo['FileVersionLS'] / 65536,
                fixedInfo['FileVersionLS'] % 65536)

        # \VarFileInfo\Translation returns list of available (language, codepage)
        # pairs that can be used to retreive string info. We are using only the first pair.
        lang, codepage = win32api.GetFileVersionInfo(fname, '\\VarFileInfo\\Translation')[0]

        # any other must be of the form \StringfileInfo\%04X%04X\parm_name, middle
        # two are language/codepage pair returned from above

        strInfo = {}
        for propName in propNames:
            strInfoPath = u'\\StringFileInfo\\%04X%04X\\%s' % (lang, codepage, propName)
            ## print str_info
            strInfo[propName] = win32api.GetFileVersionInfo(fname, strInfoPath)

        props['StringFileInfo'] = strInfo
    except:
        pass

    return props



import csv
import os
import json
from requests.auth import HTTPBasicAuth
from pmc_automation_tools.common.exceptions import PlexResponseError
from typing import Literal
from abc import ABC, abstractmethod
from requests.adapters import HTTPAdapter
from urllib3 import PoolManager
from urllib3.util.ssl_ import create_urllib3_context

"""
Base datasource class
    Shared functions
        set_auth - Done

Base input class
    shared functions
        pop_inputs - done
        update_query_string - Rename to something like update_input_parameters? - done
        purge_empty - done

Base response class
    shared functions
        save_response_csv - will need to update the UX data source to properly format the data
    shared attributes
        data_source_key - Doesn't exist for new API responses


"""
TYPE_VALUES = ['classic', 'ux', 'api']
RETRY_COUNT = 10
BACKOFF = 0.5
RETRY_STATUSES = [500, 502, 503, 504]

class CustomSslContextHTTPAdapter(HTTPAdapter):
    """"Transport adapter" that allows us to use a custom ssl context object with the requests."""
    def init_poolmanager(self, connections, maxsize, block=False):
        ctx = create_urllib3_context()
        ctx.load_default_certs()
        ctx.options |= 0x4  # ssl.OP_LEGACY_SERVER_CONNECT
        self.poolmanager = PoolManager(ssl_context=ctx)


class DataSourceInput(ABC):
    """
    """
    def __init__(self, api_id: str, type: Literal['classic', 'ux', 'api'], *args, **kwargs):
        self.__api_id__ = str(api_id)
        self.__refresh_query__ = True

        if not type.lower() in TYPE_VALUES:
            raise ValueError(f"{type(self).__name__} type must be one of {TYPE_VALUES}. Received '{type}'.")
        self.__datasource_type__ = type
        for key, value in kwargs.items():
            setattr(self, key, value)
        if kwargs.get('json'):
            self.__refresh_query__ = False
            self._query_string = kwargs['json']
        else:
            self._update_input_parameters()


    def __setattr__(self, name, value):
        self.__dict__[name] = value
        if not name.startswith('_') and self.__refresh_query__:
            self._update_input_parameters()

    @abstractmethod
    def _update_input_parameters(self):...


    def pop_inputs(self, *args, **kwargs):
        """
        Will remove attributes from the class that are not needed.

        Parameters:
        - *args: Any specific input name will be removed from the class
        - **kwargs: Can allow for keeping specific attributes when passed as a list using the "keep" kwarg.
        """
        if 'keep' in kwargs.keys():
        # if kwargs.get('keep'): # Can't use this method because an empty list is expected and evaluates false here.
            entries_to_remove = [key for key in self.__dict__.keys() if key not in kwargs['keep'] and not key.startswith('_')]
            for attr in entries_to_remove:
                vars(self).pop(attr, None)
        for attr in args:
            if attr.startswith('_'):
                continue
            vars(self).pop(attr, None)
        self._update_input_parameters()


    def purge_empty(self):
        """
        Removes any None type attributes from the object.

        These can cause issues if the input is not nullable.
        """
        purge_attrs = []
        for y in vars(self).keys():
            if getattr(self, y) is None:
                purge_attrs.append(y)
        for y in purge_attrs:
            self.pop_inputs(y)

class DataSource(ABC):
    def __init__(self, auth: HTTPBasicAuth|str,
                       test_db: bool = True,
                       pcn_config_file: str='resources/pcn_config.json',
                       type: Literal['classic', 'ux', 'api']='ux',
                       **kwargs):
        """
        Parameters:

        - auth: HTTPBasicAuth | str
            - HTTPBasicAuth object
            - API Key as a string
            - PCN Reference key for getting the username/password stored in a json config file.
            
        - test_db: bool, optional
            - Use test or production database
        
        - pcn_config_file: str, optional
            - Path to JSON file containing username/password credentials for HTTPBasicAuth connections.
        """
        
        self._test_db = test_db
        self._pcn_config_file = pcn_config_file
        self.__datasource_type__ = type
        self._auth = self.set_auth(kwargs.get('pcn', auth))


    def _check_api_key(self, input_str: str) -> bool:
        return len(input_str)==32 and input_str.isalnum() and self.__datasource_type__ == 'api'
    

    def set_auth(self, key:HTTPBasicAuth|str):
        """
        sets authentication for API calls.

        Parameters:
        - key:
            - HTTPBasicAuth object
            - API Key as a string
            - PCN Reference key for getting the username/password in a json config file.
        
        If not sending an API Key or HTTPBasicAuth object, Expects a JSON file which holds the webservice credentials.
            .. code-block:: json
                {
                    "PCN_REF":{
                        "api_user":"Username@plex.com",
                        "api_pass":"password"
                    },
                    "PCN_2_REF":{
                        "api_user":"Username2@plex.com",
                        "api_pass":"password2"
                    }
                }
        """
        if isinstance(key, HTTPBasicAuth) or self._check_api_key(key):# or key is None:
            return key
        if not os.path.exists(self._pcn_config_file):
            print(f'PCN config file "{self._pcn_config_file}" missing. Create one or enter your credentials now.')
            username = input('Webservice username:')
            password = input('Webservice password:')
        else:
            with open(self._pcn_config_file, 'r', encoding='utf-8') as c:
                self.launch_pcn_dict = json.load(c)
            username = self.launch_pcn_dict[key]['api_user']
            password = self.launch_pcn_dict[key]['api_pass']
        return HTTPBasicAuth(username, password)
    
    @abstractmethod
    def call_data_source(self):...



class DataSourceResponse(ABC):
    def __init__(self, api_id, **kwargs):
        self.__api_id__ = api_id
        for key, value in kwargs.items():
            setattr(self, key, value)


    @abstractmethod
    def _format_response(self):...

    def save_csv(self, out_file):
        """
        Save the response object to a provided CSV file.
        """
        if not getattr(self, '_transformed_data', []):
            raise PlexResponseError(f'{type(self).__name__} has no transformed data to save.')
        with open(out_file, 'w+', encoding='utf-8') as f:
            c = csv.DictWriter(f, fieldnames=self._transformed_data[0].keys(), lineterminator='\n')
            c.writeheader()
            c.writerows(self._transformed_data)
    
    
    def save_json(self, out_file):
        """
        Save the response object to a provided JSON file.
        """
        if not getattr(self, '_transformed_data', []):
            raise PlexResponseError(f'{type(self).__name__} has no transformed data to save.')
        with open(out_file, 'w+', encoding='utf-8') as f:
            f.write(json.dumps(self._transformed_data, indent=4))


    def get_response_attribute(self, attribute, preserve_list=False, **kwargs) -> list | str:
        """
        Extract the attribute from the formatted data in the response.

        Parameters:

        - attribute: attribute name from the response to return
        - preserve_list: Pass true to retain a list of attributes even if a single item is found.
        - kwargs: arbitrary number of attribute=value filters to use when searching for a specific attribute to return.

        Returns:

        - attribute(s) matching the criteria
        """
        if not kwargs:
            attr_list = [item.get(attribute) for item in self._transformed_data]
        else:
            filter_criteria = kwargs
            attr_list = []
            for item in self._transformed_data:
                if all(item.get(k) == v for k, v in filter_criteria.items()):
                    attr_list.append(item.get(attribute))
        return attr_list[0] if len(attr_list) == 1 and not preserve_list else attr_list
    get_attribute = get_response_attribute